(* Simple Prover Directly Based on Focusing *)
(* Author: Frank Pfenning *)

(* This prover uses success continuations directly *)
(*
   This is overkill for the propositional case without
   quantifiers (unless we are interested in enumerating
   solutions), since conjunctive subgoals are always
   independent.  When quantifiers are added, however,
   they become necessary.
*)

structure Focus2 :> PROVER =
struct

  datatype Reason =
    NoProof				(* no proof exists *)
  | Bound				(* bound exceeded *)

  datatype Result =
    Success of int			(* level at which found *)
  | Failure of Reason			(* no proof exists *)

  datatype Res =
    Succeeded
  | Failed       			(* no proof exists *)

  (* assume (G, A) = G,A *)
  (* since we do not yet build prove terms, hyp is no named *)
  fun assume (G, A) = C.Decl (G, ("", A))

  val boundExceeded = ref false

  (* Focused Search *)
  (* In all functions, n >= 0 is distance remaining before
     search is cut off

     A ::= P | A1 & A2 | T | A1 => A2 | A1 v A2 | F
     L ::= P | A1 => A2
     R ::= P | A1 v A2 | F

     G ::= . | G, A
     D ::= . | D, L
  *)

  (* searchR (n, D, G, A) searches D; G => A; . *)
  fun searchR (0, D, G, A, sc) = (boundExceeded := true; ())
    | searchR (n, D, G, A, sc) = (* n > 0 *)
        activeRight (n, D, G, A, sc)

  (* searchL (n, D, G, A) searches D; G => .; R *)
  and searchL (0, D, G, R, sc) = (boundExceeded := true; ())
    | searchL (n, D, G, R, sc) = (* n > 0 *)
        activeLeft (n, D, G, R, sc)

  (* activeRight (n, D, G, A)
     D; G => A; .  applies active right rules to A *)
  and activeRight (n, D, G, P.And (A, B), sc) =
        activeRight (n, D, G, A,
		     fn () => activeRight (n, D, G, B, sc))
    | activeRight (n, D, G, P.True, sc) =
	sc ()
    | activeRight (n, D, G, P.Implies (A, B), sc) =
	activeRight (n, D, assume (G, A), B, sc)
    | activeRight (n, D, G, A, sc) = (* A = P | A v B | F *)
	activeLeft (n, D, G, A, sc)

  (* activeLeft (n, D, G, R)
     D; G => .; R  applies active left rules to G *)
  and activeLeft (n, D, C.Decl (G, (u, P.And (A, B))), R, sc) =
        activeLeft (n, D, assume (assume (G, A), B), R, sc)
    | activeLeft (n, D, C.Decl (G, (u, P.True)), R, sc) =
	activeLeft (n, D, G, R, sc)
    | activeLeft (n, D, C.Decl (G, (u, P.Or (A, B))), R, sc) =
	activeLeft (n, D, assume (G, A), R,
		    fn () => activeLeft (n, D, assume (G, B), R, sc))
    | activeLeft (n, D, C.Decl (G, (u, P.False)), R, sc) =
	sc ()
    | activeLeft (n, D, C.Decl (G, (u, L)), R, sc) = (* L = P | A => B *)
	activeLeft (n, assume (D, L), G, R, sc)
    (* next to cases transition to focusing phase *)
    | activeLeft (n, D, C.Null, R as P.Atom(P), sc) =
	pickLeft (n, D, D, R, sc)
    | activeLeft (n, D, C.Null, R, sc) = (* R = A v B | F *)
	(focusRight (n, D, R, sc);
	 pickLeft (n, D, D, R, sc))

  (* focusRight (n, D, A)
     D; . >> A; .  applies focus right rules to A *)
  and focusRight (n, D, P.Or (A, B), sc) =
        (focusRight (n, D, A, sc);
	 focusRight (n, D, B, sc))
    | focusRight (n, D, P.False, sc) =
	()
    | focusRight (n, D, A, sc) = (* A = P | A1 & A2 | T | A1 => A2 *)
	searchR (n-1, D, C.Null, A, sc)

  (* pickLeft (n, D, D', R) where D' is prefix of D
     D; . => .; R  picks left focus formula from D' *)
  and pickLeft (n, D, C.Decl (D', (u, L)), R, sc) =
        (focusLeft (n, D, L, R, sc);
	 pickLeft (n, D, D', R, sc))
    | pickLeft (n, D, C.Null, R, sc) =
	()

  (* focusLeft (n, D, A, R)
     D; A >> .; R  applies focus left rules to A *)
  and focusLeft (n, D, L as P.Atom(Q), R, sc) =
      if P.eq (L, R) then sc ()
      else ()
    | focusLeft (n, D, P.Implies(A, B), R, sc) =
	focusLeft (n, D, B, R,
		   fn () => searchR (n-1, D, C.Null, A, sc))
    | focusLeft (n, D, P.And (A, B), R, sc) =
	(focusLeft (n, D, A, R, sc);
	 focusLeft (n, D, B, R, sc))
    | focusLeft (n, D, P.True, R, sc) =
	()
    | focusLeft (n, D, A, R, sc) = (* A = A1 v A2 | F *)
	searchL (n-1, D, assume (C.Null, A), R, sc)

  (* prove (n, G, A) = r *)
  (* Uses iterative deepening to search for
     G => A, represented by .; G => A; .
     bounded by n.  Bound is decreased along every
     branch whenever search on a new sequent
     D; . => A; . or D; . => .; R is initiated.

     Strategy:
     1. Right active rules, until passive
     2. Left active rules in context order, until empty
     3. Try focus right, until failure
     4. Try focus left, in context order, until failure
  *)
  fun prove (bound, G, A) =
      let
	exception Done of int
        fun iterate (n) =
	    if n > bound then Failure (Bound)
	    else (boundExceeded := false;
		  searchR (n+1, C.Null, G, A,
			   fn () => raise Done (n));
		  if !boundExceeded
		    then iterate (n+1)
		  else Failure (NoProof))
      in
	iterate (0)
	handle Done (n) => Success (n)
      end

end;  (* structure Focus2 *)
